# DockAnalyse: Main function for the paper
#        "DockAnalyse: An Algorithm for Protein-Protein Interaction Analysis"
#         Delicado P., Amela I., Gmez A., Querol E., Cedano J. (2010)
#
# (1) Reading and preprocessing the data. The resulting data are stored in "new.data"
# (2) Computing the optimal "eps" for "dbscan"
# (3) Running "dbscan" with the optimal "eps"
# (4) Selecting a representer for each cluster verifying:
#      (a) it is close to the cluster center
#      (b) it has a high score value
#
# INPUT:
# file:  The name (with extension) of an ASCII file obtained as output of a 
#        software for protein-RNA docking (as "ESCHER" or "HEX").
#        The requirements of thiis file format are the following:
#            - There are several comment lines before the numerical information.
#              The number of comment lines is provided to this R program as 
#              "num.com.lin" (>=0; by default it is equal to 15, asi in ESCHER case)
#            - The numerical information starts at line "num.com.lin"+1.
#              Each line contains several fields separated by blank spaces.
#              At least 6 fields musp appear at each line:
#                  Translation coordinates: TransX TransY TransZ
#                  Rotation coordinates: RotX RotY RotZ 
#              An optional field:
#                  Some score measure (in ascending or descending order)
# 
# num.com.lin: The number of comment lines before numerical information in "file"
# pos.trans.coor: An array with length=3 indicating the position of the fields
#         including translation coordinates at each line of "file".
#         (Default: c(13,14,15), as in ESCHER case)       
# pos.rot.coor: An array with length=3 indicating the position of the fields
#         including rotation coordinates at each line of "file".
#         (Default: c(10,11,12), as in ESCHER case)       
# pos.score: An integer indicating the position of the score-like measure.
#          If it is NULL then no score-like variable is included in "file".
#         (Default: 2, as in ESCHER case) 
# ascending.score: Logical, it is TRUE when higher score values indicate better 
#         protein adjustment; it is FALSE in the opposite case.
#         (Default: TRUE, as in ESCHER case) 
# n.eps: The number of different values of epsilon that are compared in 
#         "dbscan.eps.choice" 
#         (Default: 10)
# MinPts: Value for the parameter "MinPts" in "dbscan" (Default value: MinPts=7)
# max.noise.prop: Maximum proportion of admited points in the cluster 0 of noise points.
#         This value is used to discard classifications (usually corresponding to too small values of "eps")
#         such that too many points (a proportion greater than "max.noise.prop") in the 0 cluster,
#         when computing the indexes of cluster quality for clusters 1 to k (not including cluster 0).
#         (Default value: max.noise.prop=1, that is all classifications are consider to compute the score of
#         a given value of "eps).
# exclude.noise.cluster: When looking for the optimal "epsilon" for "dbscan" 
#         (with function "dbscan.eps.choice") the noise points are used 
#         (if  "exclude.noise.cluster" is FALSE) or not (if  "exclude.noise.cluster" is TRUE).
#         Default: FALSE.
# d.power: The distance between angles is the absolute value of the 
#         difference angle to the power "d.power". (Default: d.power=1)
# marker: Array of integer numbers (between 1 and the number of points listed in "file").
#         If "marker" is not NULL, then the points with index in "marker" are 
#         highlighted in plots. (Default: marker=NULL)
# center: The center of clusters can be computed as the mean of the cluster 
#         (if center=="mean") or as the mode (if center=="mode"). Default: "mean".
# span.mode: Proportion of data used to compute the mode (default: 0.25).
# plot.true: Logical, for drawing or not plots during execution (Default: FALSE)
# print.info: Logical, for printing or not information during execution (Default: FALSE)
#
# OUTPUT: A list with elements:
# repres: list of representer for each cluster.
# clust.dbs.opt: output of dbscan for the optimal epsilon
# out.opt.eps: output of dbscan.eps.choice
# data.proc: output of preprocessing
#
# SIDE EFFECTS: 
# A file (with the same format as "file") containing the representer points is created.
# Its name is paste("repres_",file,sep="")
# 
DockAnalyse <- function(file, num.com.lin=15,
                           pos.trans.coor=c(13,14,15),pos.rot.coor=c(10,11,12),
                           pos.score=2, ascending.score=TRUE,
                           n.eps=15, MinPts=7, max.eps=1, max.noise.prop=0.5,
                           exclude.noise.cluster=FALSE, d.power=1,col.name=1,marker=NULL,radians=FALSE,
                           center=c("mean","mode"), span.mode=.25,
                           plot.true=FALSE, print.info=FALSE, outOrig=2){
  require(fpc)
  require(ade4)
  require(clusterSim)
  
  if(!exists("preprocessing",mode="function")) source("preprocessing.R")
  if(!exists("dbscan.eps.choice",mode="function")) source("dbscan.eps.choice.R")
  
  # Reading and preprocessing the data. The resulting data are stored in "new.data"
  data.proc <- preprocessing(file=file, num.com.lin=num.com.lin,
                             pos.trans.coor=pos.trans.coor,
                             pos.rot.coor=pos.rot.coor,
                             pos.score=pos.score,
				     MinPts=MinPts,
                             radians=radians,
				     max.eps=max.eps,
                             ascending.score=ascending.score,
                             d.power=d.power, header=FALSE)

  if (print.info) cat("\n k.MDS = ", data.proc$k.MDS, ", prop.k.MDS = ", data.proc$prop.k.MDS,"\n\n")
  
  new.data <- data.proc$new.data

  dock.score <- data.proc$dock.score


  ##### We work with "new.data", a data.frame with:
  # columns 1 to 5: principal coordinates obtaind by MDS from angle information
  # columns 6 to 8: original data TransX, TransY, TransZ
  # both column groups have the same total variance (trace of covariance matrix).
  
  # Computing the optimal "eps" for "dbscan"
  #  
  out.opt.eps <- dbscan.eps.choice(x=new.data,
                      n.eps=n.eps, 
                      MinPts=MinPts, 
			    max.eps=max.eps,
                      max.noise.prop=max.noise.prop,
                      exclude.noise.cluster=exclude.noise.cluster,
                      print.info=print.info)
  
  # Running "dbscan" with the optimal "eps"
  if (plot.true) dev.set()   
  clust.dbs.opt <- dbscan(new.data, eps=out.opt.eps$opt.eps, 
                          MinPts = out.opt.eps$MinPts, showplot = plot.true)

  if (print.info) summary(as.factor(out.opt.eps$clust.dbs.opt$cluster))
  
  
  
  # Selecting a representer for each cluster verifying:
  #  (a) it is close to the cluster center
  #  (b) it has a high score value
  #
  cl.opt <- clust.dbs.opt$cluster
  ncl <- max(cl.opt) # cluster 0 is not included

  write.table(cl.opt, file= paste("Clusters_",file,sep=""),
              row.names=FALSE,quote=FALSE,sep=" ")

  if (plot.true) {dev.set();par(mfrow=c(round(sqrt(ncl)+0.5),round(sqrt(ncl)+0.5)))}
  cluster.size <- rep(0,ncl)
  repres <- rep(0,ncl)
  
  if (print.info) cat("\n")
  
  for (i in 1:ncl){
     cl.i <- which(cl.opt==i)
     ni <- sum(cl.opt==i)
     cluster.size[i] <- ni

     smean <- mean(dock.score[which(cl.opt==i)])

     if (center=="mean"){  
        mean.cl.i <- apply(dock.score[cl.i,],2,mean)         
     }else{ # mode
     # The cluster mode is calulated as the point having minimum distance to
     # their k-th nearest neighboor, where k= floor( ni * span.mode )
     # ni is the number of elements in the cluster
     # span.mode is the proportion of data used to compute the mode (default: 0.25).
        d.cl.i <- as.matrix(dist(new.data[cl.i,]))
        d.sorted <- apply(d.cl.i,2,sort)
        i.star <- which.min(floor(ni*span.mode))
        mean.cl.i <- new.data[cl.i[i.star],]
     }
 

     dmean <- sqrt( apply( (new.data[cl.i,]-matrix(mean.cl.i,nrow=ni,ncol=dim(new.data)[2],byrow=TRUE))^2,1,sum))

     cl.i.centro <- cl.i[dmean<quantile(dmean,.25)] 
     if(ascending.score){
        repres[i] <- cl.i.centro[ which.max(dock.score[cl.i.centro]) ]
     }else{
        repres[i] <- cl.i.centro[ which.min(dock.score[cl.i.centro]) ]
     }
     if (plot.true){
        plot(dmean,dock.score[cl.i],main=paste("cluster ",i," (",ni," points)",sep=""))
        points(sqrt(sum((new.data[repres[i],]-mean.cl.i)^2)),dock.score[repres[i]],pch=16,col=2)
        if (any(marker %in% cl.i)){
           marker.cl.i <- intersect( marker, cl.i )
           points(sqrt( apply( (new.data[marker.cl.i,]-matrix(mean.cl.i,nrow=length(marker.cl.i),ncol=dim(new.data)[2],byrow=TRUE))^2,1,sum)), dock.score[marker.cl.i],pch=16,col=1)         
           text(sqrt( apply( (new.data[marker.cl.i,]-matrix(mean.cl.i,nrow=length(marker.cl.i),ncol=dim(new.data)[2],byrow=TRUE))^2,1,sum)), dock.score[marker.cl.i], marker.cl.i,col=4)
        }else{
           marker.cl.i <- intersect( repres, cl.i )
           points(sqrt( apply( (new.data[marker.cl.i,]-matrix(mean.cl.i,nrow=length(marker.cl.i),ncol=dim(new.data)[2],byrow=TRUE))^2,1,sum)), dock.score[marker.cl.i],pch=16,col=1)         
           text(sqrt( apply( (new.data[marker.cl.i,]-matrix(mean.cl.i,nrow=length(marker.cl.i),ncol=dim(new.data)[2],byrow=TRUE))^2,1,sum)), dock.score[marker.cl.i], marker.cl.i,col=4)
	  }
     }
     if (print.info) cat("Cluster ", i, " (",ni," points)"," Representative Line = ", repres[i],"Energy:",dock.score[repres[i]],";  Cluster Mean Energy:", smean, "\n")
  }
  par(mfrow=c(1,1))

  ch1 <- rep(" -> ", ncl)
  ch2 <- rep("     (", ncl)
  ch3 <- rep(")", ncl)
  
  resumen_repres <- data.frame(1:ncl, ch1, repres,ch2, cluster.size, ch3)
  names(resumen_repres) <- c("Cluster","->","Representative Line:","(","cluster size",")")
  write.table(resumen_repres, file= paste("resumen_repres_",file,sep=""),
              row.names=FALSE,quote=FALSE,sep=" ")

  if (plot.true){
    dev.set()
    par(mfrow=c(2,1))
    pca <- princomp(new.data,cor=FALSE)
    plot(out.opt.eps$clust.dbs.opt,pca$scores)
    if (!is.null(marker)){
       points(pca$scores[marker,1],pca$scores[marker,2], pch=16, cex=2, col=1)
       text(pca$scores[marker,1],pca$scores[marker,2], marker,col=4)
    }else{
       points(pca$scores[repres,1],pca$scores[repres,2], pch=16, cex=2, col=1)
       text(pca$scores[repres,1],pca$scores[repres,2], repres,col=4)
    }
    boxplot(dock.score~clust.dbs.opt$cluster)
    
    dev.set()
    par(mfrow=c(1,1))
    plot(as.data.frame(cbind(new.data, pca$scores[,1:2])),col=clust.dbs.opt$cluster+1)
    
    # more graphics (at least 3 graphical windows must be open)
    plot.dbscan.eps.choice(out.opt.eps,pca,marker)
  }

if(outOrig == 0){

  # Writing the representer points file with the same format and the same comment lines as "file"
  head.sol <- read.table(file=file,header=FALSE,nrows=num.com.lin,
                         quote="",sep="*",colClasses = "character",
                         comment.char="",blank.lines.skip = FALSE)
  head.sol[[1]][9]<- paste("Solutions.........: ",ncl,sep="") # for ESCHER output files
  write.table(head.sol,file=paste("repres_",file,sep=""),col.names=FALSE,row.names=FALSE,quote=FALSE) 
  data.sol <- read.table(file=file,header=FALSE,skip=num.com.lin,quote="",sep="*",colClasses = "character")
  write.table(data.sol[repres,],file=paste("repres_",file,sep=""),col.names=FALSE,row.names=FALSE,quote=FALSE, append=TRUE)
  write.table("#END",file=paste("repres_",file,sep=""),col.names=FALSE,row.names=FALSE,quote=FALSE, append=TRUE) # for ESCHER output files
}

if(outOrig == 1){

  # Writing the representer points file with the same format and the same comment lines as "file"
  head.sol <- read.table(file=file,header=FALSE,nrows=num.com.lin,
                         quote="",sep="*",colClasses = "character",
                         comment.char="",blank.lines.skip = FALSE)
  write.table(head.sol,file=paste("repres_",file,sep=""),col.names=FALSE,row.names=FALSE,quote=FALSE) 
  data.sol <- read.table(file=file,header=FALSE,skip=num.com.lin,quote="",sep="*",colClasses = "character")
  write.table(data.sol[repres,],file=paste("repres_",file,sep=""),col.names=FALSE,row.names=FALSE,quote=FALSE, append=TRUE)

}

if(outOrig == 2){

  ch0 <- rep("frodockview ", ncl)
  ch1 <- rep(SolutionsFile, ncl)
  ch2 <- rep(" -r ", ncl)
  ch3 <- rep(" -p ", ncl)
  ch4 <- rep("_Sol.pdb", ncl)
  
  resumen_repres <- data.frame(ch0, ch1, ch2, repres, ch3, repres, ch4)
# Writing the representer for extracting with Frodockview
  write.table(resumen_repres, file= paste("resumen_repres_","extraer.sh",sep=""),
              row.names=FALSE,,col.names=FALSE,quote=FALSE,sep="")



#  write.table(head.sol,file=paste("repres_",file,sep=""),col.names=FALSE,row.names=FALSE,quote=FALSE) 
  data.sol <- read.table(file=file,header=FALSE,skip=num.com.lin,quote="",sep="*",colClasses = "character")
  write.table(data.sol[repres,],file=paste("repres_",file,sep=""),col.names=FALSE,row.names=FALSE,quote=FALSE, append=TRUE)
 
}


  return(list(repres=repres, 
              clust.dbs.opt=clust.dbs.opt, 
              out.opt.eps=out.opt.eps, 
              data.proc=data.proc))
}