# dbscan.eps.choice Automatic choice of "eps" parameter in "dbscan" clustering
# 
# Input: 
#  x, d: x is a (n x p) data matrix, d is a distance object (equivalent to a (n x n) distance matrix
#        where n is the number of objects in the sample and p is the number of variables observed in each object).
#        Data are provided to the functions in the form of either x or d (or both).
#        If x is provided (and not d), then by default d=dist(x) is the Euclidean distance beetween rows of x.
#        If d is provided (and not x), then x is the principal coordinates matrix obtained from d by MDS.
#        (Default values: x=NULL,d=NULL. One of them must be probided).
#
# v.eps: array with the candidate values of parameter "eps" for "dbscan".
#        If v.eps==NULL, then a v.eps is build with length n.eps,
#        with values comming from the Q3 of the arc lengths in the MST
#        to the exponential value of the external barrier of outliers 
#        for the set of logs of the arc lengths in the MST, computed as Q3_log + 3*(Q3_log-Q1_log). 
#        So the maximum value for v.eps is exp(Q3_log + 3*(Q3_log-Q1_log)).
#        (Default value: v.eps=NULL)
# n.eps: If v.eps==NULL n.eps indicates the length of the internally buid v.eps. 
#        If v.eps is not NULL, the value of n.eps is internally computed as length(v.eps) 
#        (Default value: n.eps=NULL)
# MinPts: Value for the parameter "MinPts" in "dbscan" (Default value: MinPts=7)
# max.noise.prop: Maximum proportion of admited points in the cluster 0 of noise points.
#        This value is used to discard classifications (usually corresponding to too small values of "eps")
#        such that too many points (a proportion greater than "max.noise.prop") in the 0 cluster,
#        when computing the indexes of cluster quality for clusters 1 to k (not including cluster 0).
#        (Default value: max.noise.prop=1, that is all classifications are consider to compute the score of
#         a given value of "eps).
# exclude.noise.cluster: When looking for the optimal "epsilon" for "dbscan" 
#         the noise points are used (if  "exclude.noise.cluster" is FALSE) 
#         or not (if  "exclude.noise.cluster" is TRUE). Default: FALSE.
# dbscan.for.opt.eps: If TRUE the dbscan for "opt.eps" is computed again. (Default=TRUE).
# print.info: Logical, for printing or not information during execution (Default: FALSE)
#
# Output: A list with the following elements:
#      v.eps: array with the examinated values of parameter "eps" for "dbscan".
#     MinPts: Value used for the parameter "MinPts" in "dbscan"
#     n.clus: number of clusters obtained by "dbscan" for each "eps" in "v.eps".
# index.clus: matrix (length(v.eps) x 8) with the vbalues of the quality indexes corresponding to 
#             clusters obtained by "dbscan" for each "eps" in "v.eps". 
#             Each row of "index.clus" corresponds to a value of "eps" in "v.eps". 
#             Each column of "index.clus" corresponds to a index:
#             "-1*Davies-Bouldin", "Calinski-Harabasz", "-1*Hubert-Levine", "Silhouette",
#             "-1*DB no noise", "CH no noise", "-1*HL no noise", "S no noise"
#             Indexes DB and HL has been changed the sign in order to have haigh values for good clusterings.
#             The first 4 indexes are calculated using all the clusters (including cluster 0: the noise cluster).
#             The last 4 indexes are calculated using only true clusters (cluster 0 is not used).
#   clus.eps: list with the clusters obtained by "dbscan" for each "eps" in "v.eps".
# n.no.noise: Number of points in the the cluster 0 of noise points.
#          n: Total number of points in the sample.
#  score.eps: Score calculated for each "eps" in "v.eps" from values of "index.clus". 
#             Given an index, the corresponding column of "index.clus" is ranked and each "eps" in "v.eps"
#             is scored with their rank in thois ranking (greater ranks mean better clustering).
#             Then the average over the eigth indexes is calculated for each "eps" in "v.eps"
#             and divided by the total number of candidates "eps" in "v.eps".
#             This value is the "score.eps" corresponding to each "eps" in "v.eps".
#  k.opt.eps: The position of the best "eps" in "v.eps" (best means highest value of "score.eps").
#    opt.eps: The best "eps" in "v.eps" (best means highest value of "score.eps").
# score.opt.eps: Value of "score.eps" for the best "eps" in "v.eps".
# clust.dbs.opt.x: Clusters obtained by "dbscan" for the best "eps" in "v.eps".
#
dbscan.eps.choice <- function(x=NULL, d=NULL, v.eps=NULL, n.eps=NULL, MinPts=7, 
                              max.noise.prop=1, exclude.noise.cluster=TRUE,
                              max.eps=1,dbscan.for.opt.eps=TRUE,
                              print.info=TRUE){
   if ((is.null(x))&(is.null(d))) stop("x or d must be supplied")
   if ((is.null(v.eps))&(is.null(n.eps))) stop("x or d must be supplied")

   if (is.null(d)) d <- dist(x)
   if (is.null(x)){
      mdsx <- cmdscale(as.dist(d), k=20, eig=TRUE)
      k <- min( sum(mdsx$eig>0), sum( cumsum(abs(mdsx$eig))/sum(abs(mdsx$eig))<=.9))
      x <- mdsx$points[,1:k]
   }
   n <- dim(x)[1]
   
   if (!is.null(v.eps)){
      n.eps <- length(v.eps)
   }else{
      # Computing the MST (minimuim spanning tree)
      mst <- mstree(d)
      dmst <- apply(mst, 1, function(x,D){return(c(D[x[1],x[2]]))},as.matrix(d))
      maxdmst <- max(dmst)
      ldmst <- log(dmst)
      Q1 <- quantile(ldmst,.25)
      Q3 <- quantile(ldmst,.75)
      int.bar <- exp(Q3 + 1.5*(Q3-Q1))
      ext.bar <- exp(Q3 + 3*(Q3-Q1))
      #outliers <- dmst>int.bar
      #extr.outliers <- dmst>ext.bar

      #eps1 <- min(int.bar,maxdmst)
      eps1 <- quantile(dmst,.75)
      eps2 <- min(ext.bar,maxdmst)
      if (eps2==eps1) warning(" * * * Warning: eps1 is equal to eps2 !!! * * * ")


      print(paste("Epsilon 1 = ",eps1,";"," Epsilon 2 = ",eps2," n.eps = ", n.eps))

      v.eps <- seq(eps1,eps2,length=n.eps)

     
   }


   index.clus <- array(0,dim=c(n.eps,8))
#   index.clus <- array(-Inf,dim=c(n.eps,8))
   colnames(index.clus)<-c("-1*Davies-Bouldin","Calinski-Harabasz","-1*Hubert-Levine","Silhouette",
                        "-1*DB no noise", "CH no noise", "-1*HL no noise", "S no noise")


   n.clus <- v.eps*0
   n.no.noise <- v.eps*0
   clus.eps <- as.list(rep(0,n.eps))
   
   for (i in 1:n.eps){
    clust <- dbscan(x, eps=v.eps[i], MinPts = MinPts)
    if(v.eps[i]<max.eps){
      if (print.info){
        cat("\n","i= ",i,", eps= ",v.eps[i],"\n",sep="")
        print(summary(as.factor(clust$cluster)))
      }
      clus.eps[[i]] <- clust$cluster
      n.clus[i] <- max(clust$cluster)
      n.no.noise[i] <- sum(clust$cluster>0)
      I.clus <- clust$cluster>0
   
      index.clus[i,1] <- -index.DB(x,clust$cluster+1)$DB
      index.clus[i,2] <- index.G1(x,clust$cluster+1)
      index.clus[i,3] <- -index.G3(d,clust$cluster+1)
      index.clus[i,4] <- index.S(d,clust$cluster+1)
   
      if ( (1-n.no.noise[i]/n) <= max.noise.prop ){
         index.clus[i,4+3] <- -index.G3(as.dist(as.matrix(d)[I.clus,I.clus]),clust$cluster[I.clus])
         if (n.clus[i]>1){ 
            index.clus[i,4+1] <- -index.DB(x[I.clus,],clust$cluster[I.clus])$DB
            index.clus[i,4+2] <- index.G1(x[I.clus,],clust$cluster[I.clus])
            index.clus[i,4+4] <- index.S(as.dist(as.matrix(d)[I.clus,I.clus]),clust$cluster[I.clus])
         }else{
            index.clus[i,4+1] <- -Inf
            index.clus[i,4+2] <- 0
            index.clus[i,4+4] <- 0
         }
      }
     }
   }

   if (exclude.noise.cluster){
      score.eps <- apply(apply(index.clus,2,rank),1,mean)/n.eps
   }else{
      score.eps <- apply(apply(index.clus[,5:8],2,rank),1,mean)/n.eps
   }
   k.opt.eps <- which.max(score.eps)
   opt.eps <- v.eps[k.opt.eps]
   score.opt.eps <- score.eps[k.opt.eps]

   clust.dbs.opt.x <- NULL
   if (dbscan.for.opt.eps){
      if (print.info){

      print(paste("Optimal epsilon= ",opt.eps," (score= ",score.opt.eps," over 1)",sep=""))
         cat("\nOptimal epsilon= ",opt.eps," (score= ",score.opt.eps," over 1)","\n\n")

      }
      clust.dbs.opt.x <- dbscan(x, eps=opt.eps, MinPts = MinPts, showplot = FALSE)
      if (print.info){
         print(summary(as.factor(clust.dbs.opt.x$cluster)))
      }
   }

   return(list(v.eps=v.eps, MinPts=MinPts, n.clus=n.clus, index.clus=index.clus, clus.eps=clus.eps, 
               n.no.noise=n.no.noise, n=n, max.noise.prop=max.noise.prop,
               score.eps=score.eps, k.opt.eps=k.opt.eps, opt.eps=opt.eps, score.opt.eps=score.opt.eps,
               clust.dbs.opt.x=clust.dbs.opt.x)
         )
}
   
# plot.dbscan.eps.choice Graphical representation of the results of "dbscan.eps.choice"
#              
# Input: 
# output.dbscan.eps.choice: A list obtained as output of "dbscan.eps.choice".
# pca: The scores of PCA (from function princomp) of the data set "new.data" obtained with function "preprocessing"
# marker: Array of integer numbers (between 1 and the number of points listed in "file.sol").
#         If "marker" is not NULL, then the points with index in "marker" are 
#         highlighted in plots. (Default: marker=NULL)
# Output: Only graphics.
# 
# Warning: Three graphical interfaces should be open (using "windows()") before running "plot.dbscan.eps.choice".
#
plot.dbscan.eps.choice <- function(output.dbscan.eps.chioce,pca,marker){
   out <- output.dbscan.eps.chioce

   dev.set()
   par(mfrow=c(2,4))
   for (i in 1:8){
      plot(out$v.eps,out$index.clus[,i],xlab="epsilon",ylab="Index",main=colnames(out$index.clus)[i])
   }
   par(mfrow=c(1,1))
   
   dev.set()
   par(mfrow=c(1,3))
   plot(out$v.eps,out$n.clus,xlab="epsilon", ylab="Number of clusters")
   plot(out$v.eps,out$n.no.noise/out$n,xlab="epsilon", ylab="Proportion of clustered points")
   abline(h=1-out$max.noise.prop,col=2,lty=2)
   plot(out$v.eps,out$score.eps,xlab="epsilon", ylab="Summary score", main="Summarizing quality indexes")
   par(mfrow=c(1,1))

   if (!is.null(out$clust.dbs.opt.x)){
      dev.set()
      plot(out$clust.dbs.opt.x,pca$scores)
      if (!is.null(marker)){
         points(pca$scores[marker,1],pca$scores[marker,2], pch=16, cex=1, col=1)
         text(pca$scores[marker,1],pca$scores[marker,2], marker,col=4)
      }
   }
}      
