# preprocessing Pre-processing the output of "docking" in order to compute distances between angles.
#               Then MDS is done on this distance matrix and the principal coordinates are concatenated
#               to position variables in the output of "docking" in order to have a new data matrix
#               where Euclidean distances between rows has sense.
#               The weight (variability) of angle information and position information is the same in the
#               new data matrix.
#
# Input:
# file:  The name (with extension) of an ASCII file obtained as output of a 
#        software for protein-RNA docking (as "ESCHER" or "HEX").
#        The requirements of thiis file format are the following:
#            - There are several comment lines before the numerical information.
#              The number of comment lines is provided to this R program as 
#              "num.com.lin" (>=0; by default it is equal to 15, asi in ESCHER case)
#            - The numerical information starts at line "num.com.lin"+1.
#              Each line contains several fields separated by blank spaces.
#              At least 6 fields musp appear at each line:
#                  Translation coordinates: TransX TransY TransZ
#                  Rotation coordinates: RotX RotY RotZ 
#              An optional field:
#                  Some score measure (in ascending or descending order)
# num.com.lin: The number of comment lines before numerical information in "file"
# pos.trans.coor: An array with length=3 indicating the position of the fields
#         including translation coordinates at each line of "file".
#         (Default: c(13,14,15), as in ESCHER case)       
# pos.rot.coor: An array with length=3 indicating the position of the fields
#         including rotation coordinates at each line of "file".
#         (Default: c(10,11,12), as in ESCHER case)       
# pos.score: An integer indicating the position of the score-like measure.
#          If it is NULL then no score-like variable is included in "file".
#         (Default: 2, as in ESCHER case) 
# ascending.score: Logical, it is TRUE when higher score values indicate better 
#         protein adjustment; it is FALSE in the opposite case.
#         (Default: TRUE, as in ESCHER case) 
# k0.MDS.max: Number of principal coordinates initailaly computed from angle distances (Default: k0.MDS.max=6) 
#      k.MDS: Number of finally desired principal corrdinates. 
#             If "k.MDS" is NULL, then "k.MDS" is internally computed to 
#             recover a proportion "prop.MDS" of the total geometrical variablitiy 
#             associated to angle distance matrix (Default: k.MDS = NULL).
#   prop.MDS: The minimum proportion of total geometrical variablitiy 
#             associated to angle distance matrix that the first "k.MDS" principal coordinates
#             must explain. This parameter is used only if "k.MDS" is NULL. (Default: prop.MDS =.95)
#   d.power:  The distance between angles is the absolute value of the 
#             difference angle to the power "d.power". Default: d.power=1.
#   header: Parameter to be passed to "read.table". Default: header=FALSE.
#
# Output: A list with the following elements:
# dock.score: Docking score (second column in file "file").
#   new.data: The new data matrix, with dimensions n x (k.MDS+3).
#      k.MDS: The number of principal coordinates obtained from angle distances by MDS 
#             that finally are stored as the first columns of new.data
# prop.k.MDS: The proportion of total geometrical variablitiy associated to the first "k.MDS" 
#             principal coordinates obtained from angle distances by MDS.
#
preprocessing <- function(file, num.com.lin=15, 
                          pos.trans.coor=c(13,14,15),pos.rot.coor=c(10,11,12),
                          pos.score=2, ascending.score=TRUE, radians=FALSE,
	  			  max.eps=1, MinPts=7,
                          k0.MDS.max=6, k.MDS = NULL, prop.MDS =.75, 
                          d.power=1, header=FALSE){
   data.total <- read.table(file=file,header=header,quote="",sep="",skip=num.com.lin)
   
#write.table(data.total,row.names=FALSE,quote=FALSE,sep=" ")
   if (is.null(pos.score)){
       score <- data.total[,1]*0
   }else{
#       score <- data.total[,pos.score] * (2*as.integer(ascending.score)-1)
       score <- data.total[,pos.score]
    } 

# We remove non-interacting solutions
# data.total <- data.total[which(score>0),] 
# We restore the engergy value including it sign

#   score <- data.total[,pos.score] 
	
   n <- dim(data.total)[1]
   noms <- c("RotX", "RotY", "RotZ", "TransX", "TransY", "TransZ")
   
   # Reading angles
   data.angle <- matrix(0,ncol=3,nrow=n)
   if(radians){
      data.angle[,1] <- data.total[,pos.rot.coor[1]]
      data.angle[,2] <- data.total[,pos.rot.coor[2]]
      data.angle[,3] <- data.total[,pos.rot.coor[3]]
   }else{
      data.angle[,1] <- pi*data.total[,pos.rot.coor[1]]/180
      data.angle[,2] <- pi*data.total[,pos.rot.coor[2]]/180
      data.angle[,3] <- pi*data.total[,pos.rot.coor[3]]/180
   }
   # Computing distances between angles 
   # as the arc length in a unit radious sphere of the difference angle.
   #
   X <- array(data.angle[,1],dim=c(n,n))
   b_a <- X - t(X)
   d1 <- pmin(abs(b_a),abs(b_a+2*pi),abs(b_a-2*pi))^d.power
   X <- array(data.angle[,2],dim=c(n,n))
   b_a <- X - t(X)
   d2 <- pmin(abs(b_a),abs(b_a+2*pi),abs(b_a-2*pi))^d.power
   X <- array(data.angle[,3],dim=c(n,n))
   b_a <- X - t(X)
   d3 <- pmin(abs(b_a),abs(b_a+2*pi),abs(b_a-2*pi))^d.power
   D <- sqrt(d1^2+d2^2+d3^2)

   # MDS for D, the distance matrix between angles
   mds.angles <- cmdscale(as.dist(D), k=k0.MDS.max, eig=TRUE)
   if (is.null(k.MDS)){
      k <- min( sum(mds.angles$eig>0), sum( cumsum(abs(mds.angles$eig))/sum(abs(mds.angles$eig))<=prop.MDS)+1)
   }else{
      k <- min( sum(mds.angles$eig>0), k.MDS )
   }
   prop.k.MDS = sum(mds.angles$eig[1:k])/sum(abs(mds.angles$eig))
   
   # concatenating principal coordinates from angles and position variables
   new.data <- cbind(mds.angles$points[,1:k],as.matrix(data.total[,pos.trans.coor]))
   colnames(new.data) <- c(paste("MDS",1:k,sep=""),noms[4:6])

   # re-sclaing the variables in order to have the same total variance from the angle and the position information
   kk <- dim(new.data)[2]
   cscov <- cumsum(diag(cov(new.data)))
   csmds <- cscov[k]
   csresto <- cscov[kk]-csmds
   new.data[,1:k] <- new.data[,1:k]/sqrt(csmds)
   new.data[,(k+1):kk] <- new.data[,(k+1):kk]/sqrt(csresto)


   return(list(dock.score=score, new.data=new.data, k.MDS=k, prop.k.MDS=prop.k.MDS))
}
   



